
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "openglheader.h"

#include "utilities.h"
#include "meshes.h"

#define LOCAL_GROUP_SIZE 256
#define GRSIZE(a) (((a)+LOCAL_GROUP_SIZE-1)/LOCAL_GROUP_SIZE)

#define MVFBUF   mbuf[0]
#define MHEBUF   mbuf[1]
#define VCBUF    mbuf[2]

#define FHEMASK  0x01FFFFFF
#define TAGMASK  0x02000000
#define DEGMASK  0xFC000000
#define DEGSHIFT         26

#define SSB GL_SHADER_STORAGE_BUFFER

/* ////////////////////////////////////////////////////////////////////////// */
static GLuint ref_shader_id, ref_program_id;
static GLuint uvloc[16];

void LoadMeshRefinementShaders ( void )
{
  const GLchar *filename[] =
    { "md.glsl.comp" };
  const GLchar *uvname[] =
    { "prN0", "prN", "prStep", "nsattr", "inv", "inhe", "infac",
      "outnv", "outnhe", "outnfac", "stage", "invb", "inei",
      "fvf", "maxonv", "fvhe" };
  int i;

  ref_shader_id = CompileShaderFiles ( GL_COMPUTE_SHADER, 1, &filename[0] );
  ref_program_id = LinkShaderProgram ( 1, &ref_shader_id );
  for ( i = 0; i < 16; i++ )
    uvloc[i] = glGetUniformLocation ( ref_program_id, uvname[i] );
  ExitIfGLError ( "LoadMeshRefinementShaders" );
} /*LoadMeshRefinementShaders*/

void DestroyMeshRefinementShaders ( void )
{
  glUseProgram ( 0 );
  glDeleteShader ( ref_shader_id );
  glDeleteProgram ( ref_program_id );
  ExitIfGLError ( "DestroyMeshRefinementShaders" );
} /*DestroyMeshRefinementShaders*/

/* ////////////////////////////////////////////////////////////////////////// */
static void iPrefixSum ( unsigned int N0, unsigned int N )
{
  int k, m, s;

  glUniform1i ( uvloc[10], 0 );
  glUniform1i ( uvloc[0], N0 );
  glUniform1i ( uvloc[1], N );
  s = (N+2*LOCAL_GROUP_SIZE-1)/(2*LOCAL_GROUP_SIZE);
  for ( k = 0, m = N-1;  m > 0;  k++, m >>= 1 ) {
    glUniform1i ( uvloc[2], k );
    glDispatchCompute ( s, 1, 1 );
    glMemoryBarrier ( GL_SHADER_STORAGE_BARRIER_BIT );
  }
  ExitIfGLError ( "iPrefixSums" );
} /*iPrefixSum*/

/* ////////////////////////////////////////////////////////////////////////// */
static void ExecStage ( int stage, int gsize )
{
  glUniform1i ( uvloc[10], stage );
  glDispatchCompute ( gsize, 1, 1 );
  glMemoryBarrier ( GL_SHADER_STORAGE_BARRIER_BIT );
} /*ExecStage*/

char GPUmeshDoubling ( GPUmesh *inmesh, GPUmesh *outmesh )
{
  int    inv, inhe, infac, invb, inei,
         onv, onhe, onfac, fvf, maxonv;
  int    bufsize, ivgsize, ihgsize, ifgsize, ovgsize;
  GLuint auxbuf;

  glUseProgram ( ref_program_id );
  inv = inmesh->nv;  inhe = inmesh->nhe;  infac = inmesh->nfac;
  ivgsize = GRSIZE ( inv );
  ihgsize = GRSIZE ( inhe );
  ifgsize = GRSIZE ( infac );
  glBindBufferBase ( SSB, 1, inmesh->MVFBUF );
  glBindBufferBase ( SSB, 2, inmesh->MHEBUF );
  glBindBufferBase ( SSB, 3, inmesh->VCBUF );
  glUniform1i ( uvloc[3], inmesh->nsattr );
  glUniform1i ( uvloc[4], inv );
  glUniform1i ( uvloc[5], inhe );
  glUniform1i ( uvloc[6], infac );
  glUniform1i ( uvloc[14], maxonv = inhe+2*inv );
  bufsize = (3*inv + 4*inhe + infac + 3)*sizeof(GLint);
  glGenBuffers ( 1, &auxbuf );
  glBindBufferBase ( SSB, 0, auxbuf );
  glBufferData ( SSB, bufsize, NULL, GL_DYNAMIC_DRAW );

  ExecStage ( 1, ivgsize );  /* TagVertex */
  iPrefixSum ( 0, inv );
  glGetBufferSubData ( SSB, (inv-1)*sizeof(GLint), sizeof(GLint), &invb );
  inei = (inhe-invb)/2;
  onv = inhe + 2*invb;
  onhe = 8*(inei + invb);
  onfac = infac + inei + invb + inv;
  glUniform1i ( uvloc[15], onhe - 2*invb );  /* fvhe */
  if ( !ReallocGPUmesh ( outmesh, onv, onhe, onfac, inmesh->nsattr,
                         inmesh->pdim, inmesh->pofs, inmesh->nvofs,
                         inmesh->tdim, inmesh->tofs ) )
    goto failure;
  glBindBufferBase ( SSB, 4, outmesh->MVFBUF );
  glBindBufferBase ( SSB, 5, outmesh->MHEBUF );
  glBindBufferBase ( SSB, 6, outmesh->VCBUF );
  glUniform1i ( uvloc[7], onv );
  glUniform1i ( uvloc[8], onhe );
  glUniform1i ( uvloc[9], onfac );
  glBindBufferBase ( SSB, 0, auxbuf );
  ExecStage ( 2, ihgsize );  /* DSetECN */
  iPrefixSum ( maxonv, inhe+1 );
  ExecStage ( 3, ivgsize );  /* DSetVCN */
  iPrefixSum ( maxonv+inhe+1, inv+1 );
  ExecStage ( 4, ivgsize );  /* DCopyVC */
  ovgsize = GRSIZE ( onv );
  ExecStage ( 5, ovgsize );  /* DSetOVdeg */
  iPrefixSum ( 0, onv );
  ExecStage ( 6, ovgsize );  /* DSetOVfhe */
  ExecStage ( 7, ifgsize );  /* DSetWLF */
  ExecStage ( 8, ihgsize );  /* DSetEFN1 */
  iPrefixSum ( maxonv+inhe+inv+2, inhe );
  ExecStage ( 9, ihgsize );  /* DSetEFN2 */
  glBindBuffer ( GL_COPY_READ_BUFFER, inmesh->MVFBUF );
  glBindBuffer ( GL_COPY_WRITE_BUFFER, outmesh->MVFBUF );
  glCopyBufferSubData ( GL_COPY_READ_BUFFER, GL_COPY_WRITE_BUFFER,
                 inv*sizeof(GLint), onv*sizeof(GLint), infac*sizeof(GLint) );
  glUniform1i ( uvloc[11], invb );
  glUniform1i ( uvloc[12], inei );
  glUniform1i ( uvloc[13], fvf = infac+inei+invb );
  ExecStage ( 10, GRSIZE(inei+invb) );  /* DSetOMfac1  */
  ExecStage ( 11, ivgsize );  /* DSetOMfac2 */
  iPrefixSum ( 0, fvf-infac+1 );
  ExecStage ( 12, ivgsize );  /* DSetOMfac3 */
  ExecStage ( 13, ihgsize );  /* DBindNewhe1 */
  iPrefixSum ( 0, inhe );
  ExecStage ( 14, ihgsize );  /* DBindNewhe2 */
  ExecStage ( 15, ihgsize );  /* DBindNewhe3 */
  ExecStage ( 16, ifgsize );  /* DSetIFDeg */
  iPrefixSum ( maxonv+3*inhe+inv+3, infac );
  ExecStage ( 17, ifgsize );  /* DSetOMfhei1 */
  ExecStage ( 18, ihgsize );  /* DSetOMfhei2 */
  ExecStage ( 19, ivgsize );  /* DSetTgv */
  iPrefixSum ( 0/*maxonv+4*inhe+inv+infac+3*/, inv );
  ExecStage ( 20, ivgsize );  /* DSetOMfhei3 */
  glDeleteBuffers ( 1, &auxbuf );
  glUseProgram ( 0 );
  ExitIfGLError ( "GPUmeshDoubling" );
  return true;

failure:
  glDeleteBuffers ( 1, &auxbuf );
  glUseProgram ( 0 );
  return false;
} /*GPUmeshDoubling*/

char GPUmeshAveraging ( GPUmesh *inmesh, GPUmesh *outmesh )
{
  int    inv, inhe, infac, invb, onv, onhe, onfac;
  int    bufsize, bs, ivgsize, ihgsize, ifgsize;
  GLuint auxbuf;

  glUseProgram ( ref_program_id );
  inv = inmesh->nv;  inhe = inmesh->nhe;  infac = inmesh->nfac;
  ivgsize = GRSIZE ( inv );
  ihgsize = GRSIZE ( inhe );
  ifgsize = GRSIZE ( infac );

  glBindBufferBase ( SSB, 1, inmesh->MVFBUF );
  glBindBufferBase ( SSB, 2, inmesh->MHEBUF );
  glBindBufferBase ( SSB, 3, inmesh->VCBUF );
  glUniform1i ( uvloc[3], inmesh->nsattr );
  glUniform1i ( uvloc[4], inv );
  glUniform1i ( uvloc[5], inhe );
  glUniform1i ( uvloc[6], infac );
  bs = inv > infac ? inv : infac;
  bufsize = (2*inv+2*inhe+infac+bs)*sizeof(GLint);
  glGenBuffers ( 1, &auxbuf );
  glBindBufferBase ( SSB, 0, auxbuf );
  glBufferData ( SSB, bufsize, NULL, GL_DYNAMIC_DRAW );

  ExecStage ( 1, ivgsize );  /* TagVertex */
  iPrefixSum ( 0, inv );
  glGetBufferSubData ( SSB, (inv-1)*sizeof(GLint), sizeof(GLint), &invb );
  onfac = inv - invb;
  if ( invb == 0 ) {
    onv = infac;  onhe = inhe;  onfac = inv;
    ExecStage ( 21, ifgsize );  /* ASetNvi1 */
    ExecStage ( 22, ihgsize );  /* ASetNhei1 */
    ExecStage ( 23, ivgsize );  /* ASetNfi1 */
  }
  else {
    ExecStage ( 24, ifgsize );  /* ASetNvi2 */
    iPrefixSum ( 0, infac );
    iPrefixSum ( 2*infac+inhe+inv, infac );
    glGetBufferSubData ( SSB, (3*infac+inhe+inv-1)*sizeof(GLint),
                         sizeof(GLint), &onv );
    ExecStage ( 25, ivgsize );  /* ASetNfi2, true */
    iPrefixSum ( 2*infac+inhe, inv );
    ExecStage ( 26, ihgsize );  /* ASetNhei2 */
    iPrefixSum ( 2*infac, inhe );
    glGetBufferSubData ( SSB, (2*infac+inhe-1)*sizeof(GLint),
                         sizeof(GLint), &onhe );
    ExecStage ( 27, ifgsize );  /* ASetNvi2, false */
    ExecStage ( 28, ihgsize );  /* ASetNhei3 */
    ExecStage ( 29, ivgsize );  /* ASetNfi3 */
  }
  if ( !ReallocGPUmesh ( outmesh, onv, onhe, onfac, inmesh->nsattr,
                         inmesh->pdim, inmesh->pofs, inmesh->nvofs,
                         inmesh->tdim, inmesh->tofs ) )
    goto failure;
  glBindBufferBase ( SSB, 4, outmesh->MVFBUF );
  glBindBufferBase ( SSB, 5, outmesh->MHEBUF );
  glBindBufferBase ( SSB, 6, outmesh->VCBUF );
  glUniform1i ( uvloc[7], onv );
  glUniform1i ( uvloc[8], onhe );
  glUniform1i ( uvloc[9], onfac );
  ExecStage ( 30, ivgsize );  /* AClearFVd */
  ExecStage ( 31, ivgsize );  /* ASetFVd1 */
  iPrefixSum ( 2*infac+inhe+inv, inv );
  ExecStage ( 32, ivgsize );  /* ASetFVd2 */
  ExecStage ( 33, ifgsize );  /* AClearFvd */
  ExecStage ( 34, ifgsize );  /* ASetOMVert, true */
  iPrefixSum ( 2*infac+inhe+inv, infac );
  ExecStage ( 35, ifgsize );  /* ASetOMVert, false */
  ExecStage ( 36, ihgsize );  /* ABindHe */
  ExecStage ( 37, ivgsize );  /* ASetOMfacHe */
  ExecStage ( 38, ifgsize );  /* Average */
  glUseProgram ( 0 );
  glDeleteBuffers ( 1, &auxbuf );
  ExitIfGLError ( "GPUmeshAveraging" );
  return true;

failure:
  glUseProgram ( 0 );
  glDeleteBuffers ( 1, &auxbuf );
  return false;
} /*GPUMeshAveraging*/

char GPUmeshRefinement ( int n, GPUmesh *inmesh, GPUmesh *outmesh )
{
  GPUmesh mmesh, *am, *bm, *cm;
  int     i;

  if ( n < 1 )
    return false;
  memset ( &mmesh, 0, sizeof(GPUmesh) );
  if ( n & 0x01 ) { am = &mmesh;  bm = outmesh; }
             else { am = outmesh;  bm = &mmesh; }
  if ( !GPUmeshDoubling ( inmesh, am ) )
    goto failure;
  for ( i = 0; i < n; i++ ) {
    if ( !GPUmeshAveraging ( am, bm ) )
      goto failure;
    cm = am;  am = bm;  bm = cm;
  }
  glDeleteBuffers ( 3, mmesh.mbuf );
  return true;

failure:
  glDeleteBuffers ( 3, mmesh.mbuf );
  return false;
} /*GPUmeshRefinement*/

